using System;
using UnityEngine;
using UnityEngine.Experimental.Rendering;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using AssetBundles;
using System.Diagnostics;
using UnityEngine.EventSystems;
using UnityEngine.UI;

namespace Stopper {
	public class Undocker : MVRScript {

        Transform pluginUI;
        Transform modelUIcanvas;

        HashSet<MVRPluginManager> pluginManagers = new HashSet<MVRPluginManager>();
        public HashSet<MVRScript> plugins = new HashSet<MVRScript>();

        List<UnityEngine.Object> createdObjects = new List<UnityEngine.Object>();
        List<UndockHandler> handlers = new List<UndockHandler>();

        GameObject alwaysShowContainer;
        Canvas alwaysShowCanvas;

        Dictionary<string, UndockerSettings> savedSettings = new Dictionary<string, UndockerSettings>();

        bool isSessionPlugin = false;

        public class UndockerSettings {
            public Vector2 offset = Vector2.zero;
            public bool alwaysShow = false;
        }

		public override void Init() {
			isSessionPlugin = containingAtom.name == "CoreControl";         

            if(isSessionPlugin){
                SuperController.singleton.onAtomAddedHandlers += OnAtomAdded;
                SuperController.singleton.onAtomRemovedHandlers += OnAtomRemoved;
                foreach(Atom atom in SuperController.singleton.GetAtoms()){
                    OnAtomAdded(atom);
                }   
            } else {
                OnAtomAdded(containingAtom);
            }


            // Create 'always show' container
            alwaysShowContainer = new GameObject("AlwaysShowPluginUIs");
            alwaysShowContainer.transform.parent = SuperController.singleton.sceneControlUI;
            alwaysShowContainer.SetActive(true);
            createdObjects.Add(alwaysShowContainer);

            alwaysShowCanvas = alwaysShowContainer.AddComponent<Canvas>();
            alwaysShowCanvas.renderMode = RenderMode.WorldSpace;

            alwaysShowCanvas.transform.rotation = SuperController.singleton.sceneControlUI.rotation;
            alwaysShowCanvas.transform.position = new Vector3(0.0f, 0.0f, 0f);
            alwaysShowCanvas.transform.localPosition = new Vector3(0f,0f,0f);

            alwaysShowCanvas.transform.localScale = new Vector3(1f,1f,1f);
            
            SuperController.singleton.AddCanvas(alwaysShowCanvas);
            
            RectTransform mainMenuRect = SuperController.singleton.mainMenuUI.GetComponent<RectTransform>();
            RectTransform sceneControlRect = SuperController.singleton.sceneControlUI.GetComponent<RectTransform>();
            RectTransform alwaysShowRect = alwaysShowCanvas.GetComponent<RectTransform>();
            alwaysShowRect.anchorMin = new Vector2(0f,0f);
            alwaysShowRect.anchorMax = new Vector2(0f,0f);
            alwaysShowRect.offsetMin = mainMenuRect.offsetMin;
            alwaysShowRect.offsetMax = mainMenuRect.offsetMax;
            alwaysShowRect.anchoredPosition -= sceneControlRect.offsetMin;
            alwaysShowRect.anchoredPosition += new Vector2(0f,16f);
            alwaysShowRect.pivot = new Vector2(0.5f,0.5f);
            
            CanvasScaler cs = alwaysShowContainer.AddComponent<CanvasScaler>();
            cs.scaleFactor = 200.0f;
            cs.dynamicPixelsPerUnit = 1f;

            GraphicRaycaster gr = alwaysShowContainer.AddComponent<GraphicRaycaster>();
		}

        private void Start() {
            if(isSessionPlugin) {
                StartCoroutine("UpdatePluginLists");
            }
        }

        private IEnumerator UpdatePluginLists(){
            while(true) {
                plugins.RemoveWhere((p) => p == null);

                foreach(var atom in SuperController.singleton.GetAtoms()){
                    foreach(string storableId in atom.GetStorableIDs()){
                        var storable = atom.GetStorableByID(storableId);
                        if(storable is MVRScript){
                            var plugin = storable as MVRScript;
                            if(!plugins.Contains(plugin)) {
                                try {
                                    if(SetUpPlugin(plugin)) {
                                        plugins.Add(plugin);
                                    }
                                } catch (Exception e) {
                                    SuperController.LogMessage($"Error setting up undocker for {plugin}: {e}");
                                    plugins.Add(plugin);
                                }
                            }
                        }
                    }
                }
                yield return new WaitForSeconds(0.5f);
            }
        }

        private void OnAtomAdded(Atom atom) {
            MVRPluginManager pluginManager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
            pluginManagers.Add(pluginManager);
        }

         private void OnAtomRemoved(Atom atom) {
            MVRPluginManager pluginManager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
            pluginManagers.Remove(pluginManager);
        }

        private bool SetUpPlugin(MVRScript script) {
            if(script.containingAtom == null || script.containingAtom.UITransform == null) {
                return false;
            }

            var atomUICanvas = script.containingAtom.UITransform.GetComponentInChildren<Canvas>(true);
            if(atomUICanvas == null) {
                SuperController.LogMessage($"Undocker couldn't to find ui canvas for atom {script.containingAtom}");
                return true; // return true here since we don't want to try again for this atom
            }

            // Undock Button
            var undockButton = CreateButton("Drag");
            undockButton.name = "UndockButton";
            createdObjects.Add(undockButton.gameObject);
            undockButton.transform.SetParent(script.UITransform, false);

            var undockRect = undockButton.GetComponent<RectTransform>();
            undockRect.anchorMin = new Vector2(0.1f,0f);
            undockRect.anchorMax = new Vector2(0.1f,0f);
            undockRect.offsetMin = new Vector2(-225.4f, 5.0f);
            undockRect.offsetMax = new Vector2(-14.6f, 55.0f);
            undockRect.anchoredPosition = new Vector2(0,30f);
            undockRect.pivot = new Vector2(0.5f, 0.5f);

             // Main handler
            var handler = undockButton.gameObject.AddComponent<UndockHandler>();
            handlers.Add(handler);


            UndockerSettings settings = new UndockerSettings();
            string pluginId = $"{containingAtom.uid}#{script.storeId}";
            if(savedSettings.ContainsKey(pluginId)){
                settings = savedSettings[pluginId];
            } else {
                savedSettings[pluginId] = settings;
            }

            handler.Initialize(atomUICanvas.gameObject.transform, 
                alwaysShowContainer.transform, 
                script, 
                script.UITransform.GetComponent<RectTransform>(), 
                settings,
                this);

            // Always Show Toggle
            JSONStorableBool alwaysShow = new JSONStorableBool("Always Show", settings.alwaysShow);
            var alwaysShowToggle = CreateToggle(alwaysShow);
            alwaysShowToggle.name = "AlwaysShowToggle";
            createdObjects.Add(alwaysShowToggle.gameObject);
            alwaysShowToggle.transform.SetParent(script.UITransform, false);
            alwaysShowToggle.toggle.onValueChanged.AddListener((val) => {
                handler.AlwaysShow(val);
            });

            var alwaysShowRect = alwaysShowToggle.GetComponent<RectTransform>();
            alwaysShowRect.anchorMin = new Vector2(0.5f,0f);
            alwaysShowRect.anchorMax = new Vector2(0.53f,0f);
            alwaysShowRect.offsetMin = new Vector2(-225.4f, 5.0f);
            alwaysShowRect.offsetMax = new Vector2(-14.6f, 55.0f);
            alwaysShowRect.anchoredPosition = new Vector2(0,30f);
            alwaysShowRect.pivot = new Vector2(0.5f, 0.5f);
            
            // Dock Button
            var dockButton = CreateButton("Dock");
            dockButton.name = "DockButton";
            createdObjects.Add(dockButton.gameObject);
            dockButton.transform.SetParent(script.UITransform, false);

            var dockRect = dockButton.GetComponent<RectTransform>();
            dockRect.anchorMin = new Vector2(0.3f,0f);
            dockRect.anchorMax = new Vector2(0.3f,0f);
            dockRect.offsetMin = new Vector2(-225.4f, 5.0f);
            dockRect.offsetMax = new Vector2(-14.6f, 55.0f);
            dockRect.anchoredPosition = new Vector2(00f,30f);
            dockRect.pivot = new Vector2(0.5f, 0.5f);
            dockButton.button.onClick.AddListener(() =>
            {
                handler.Dock();
            });

            // Bring To Front On Click
            var bringToFront = script.UITransform.gameObject.AddComponent<BringToFrontHandler>();
            createdObjects.Add(bringToFront);

            return true;
        }
		
        private string GetPathToObject(GameObject go) {
		    if (go.transform.parent == null)
			     return "/" + go.name;
            return GetPathToObject(go.transform.parent.gameObject) + "/" + go.name;
	    }   
		
        private void OnDisable() {
            foreach(var handler in handlers) {
                handler.Dock();
            }
        }
		
		private void OnDestroy() {
            SuperController.singleton.onAtomAddedHandlers -= OnAtomAdded;
            SuperController.singleton.onAtomRemovedHandlers -= OnAtomRemoved;


            
            if(alwaysShowCanvas != null) {
                SuperController.singleton.RemoveCanvas(alwaysShowCanvas);
            }

			foreach(var o in createdObjects) {
                if(o != null) {
                    Destroy(o);
                }
            }
		}
	}

    public class UndockHandler : MonoBehaviour, IPointerDownHandler, IPointerUpHandler {
        private Transform modelUIcanvas;
        private Transform alwaysShowTransform;
        private MVRScript script;
        private RectTransform uiRectTransform;
        private Undocker.UndockerSettings settings;
        private Undocker parentPlugin;
        
        private bool undocked = false;
        private bool down;
        private Vector3 mousePos;

        private Transform originalParent;
        private Vector2 originalAnchoredPosition;

        public void Initialize(Transform modelUIcanvas, Transform alwaysShowTransform, MVRScript script, RectTransform uiRectTransform, Undocker.UndockerSettings settings, Undocker parentPlugin) {
            this.modelUIcanvas = modelUIcanvas;
            this.alwaysShowTransform = alwaysShowTransform;
            this.script = script;
            this.uiRectTransform = uiRectTransform;
            this.settings = settings;

            
            if(this.settings.offset != Vector2.zero) {
               // AlwaysShow(true);
                Undock();
                script.UITransform.gameObject.SetActive(true);
            }

            this.uiRectTransform.offsetMin += settings.offset;
            this.uiRectTransform.offsetMax += settings.offset;
            this.uiRectTransform.anchoredPosition += settings.offset;


            this.parentPlugin = parentPlugin;
        }
        
        public void Update(){
            if(down && undocked) {
                var mouseDelta = Input.mousePosition - mousePos;
                
                var deltaVector = new Vector2(mouseDelta.x, mouseDelta.y);
                uiRectTransform.offsetMin += deltaVector;
                uiRectTransform.offsetMax += deltaVector;
                uiRectTransform.anchoredPosition += deltaVector;
                
                settings.offset += deltaVector;

                mousePos = Input.mousePosition;
            }
        }

        public void OnPointerDown(PointerEventData eventData) {
            if(!undocked) {
                Undock();
            }
            
            //script.UITransform.SetAsLastSibling();
            down = true;
            mousePos = Input.mousePosition;
        }

        public void AlwaysShow(bool show) {
            settings.alwaysShow = show;
            if(undocked){
                if(show) {
                    script.UITransform.SetParent(alwaysShowTransform,false);
                } else {
                    script.UITransform.SetParent(modelUIcanvas,false);
                }
            }
        }

        public void OnPointerUp(PointerEventData eventData) {
            down = false;
        }

        
        public void Undock() {
            originalParent = script.UITransform.parent;
                
            var originalRectTransform = script.UITransform.GetComponent<RectTransform>();
            originalAnchoredPosition = originalRectTransform.anchoredPosition;

            if(settings.alwaysShow) {
                script.UITransform.SetParent(alwaysShowTransform,false);
            } else {
                script.UITransform.SetParent(modelUIcanvas,false);
            }
            undocked = true;
        }

        public void Dock() {
            if(undocked) {
                script.UITransform.SetParent(originalParent,false);

                var rectTransform = script.UITransform.GetComponent<RectTransform>();
                rectTransform.offsetMin = Vector2.zero;
                rectTransform.offsetMax = Vector2.zero;
                rectTransform.anchoredPosition = originalAnchoredPosition;
                this.settings.offset = Vector2.zero;

                script.UITransform.gameObject.SetActive(false);

                undocked = false;
            }
        }
    }

    public class BringToFrontHandler : MonoBehaviour, IPointerClickHandler {
        
        public void OnPointerClick(PointerEventData eventData) {
            transform.SetAsLastSibling();
            /*
            List<RaycastResult> raycastResults = new List<RaycastResult>();
            EventSystem.current.RaycastAll(eventData, raycastResults);
            SuperController.LogMessage($"{raycastResults.Count}");
            foreach(var result in raycastResults) {
                if(result.gameObject.GetComponent<IPointerDownHandler>() != null) {
                    SuperController.LogMessage($"{result.gameObject}");
                }
                if(result.gameObject == this.gameObject){
                    continue;
                }
                if(result.gameObject.GetComponent<IPointerDownHandler>() != null) {
                    ExecuteEvents.Execute(result.gameObject, eventData, ExecuteEvents.pointerDownHandler);
                    return;
                }
            }*/
        }
    }
}